﻿/*
 *	Source file	: mp3_13_menu.c
 *	Date		: 2019/01/12
 *	Update		: 
 */
 
#include <stdint.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <fcntl.h>
#include <sys/types.h>
#include <dirent.h>
#include <string.h>
#include <sys/ioctl.h>
#include <linux/types.h>
#include <linux/spi/spidev.h>
#include <wiringPi.h>
#include <sys/wait.h>
#include <errno.h>
#include <sys/time.h>
#include <signal.h>
#include <time.h>
#include <sys/socket.h>

#include "mp3_13.h"

// Raspberry PI からSPIを通じてPICへ送られるコマンド
#define CMD_IMAGEDATA       0xF1        // イメージデータ送信開始
#define CMD_LCDDISPON       0xF2        // LCD表示をオン(送信済みイメージデータを再表示)
#define CMD_LCDDISPOFF      0xF3        // LCD表示をオフ
#define CMD_POWERDOWN       0xF4        // 電源オフ
#define CMD_ACTIVE          0xF5		// 本PGが実行中であることをPICに通知

// 入出力のピン番号をwiringPIのピン番号で定義
// SWに使用されるwiringPI PIN は 0,4,5,7,8,9,15,16
#define PIN_SWDOWN				8
#define	PIN_SWUP				9
#define PIN_SWNEXT				5
#define PIN_SWPREV				4
#define PIN_SWSEL				7
#define PIN_SWDESEL				16
#define PIN_SWPLAY				0
#define PIN_SWOPTION			15
// SW以外のI/O
#define PIN_INP_PDOWN			21		// PICからのPOWER DOWN指示
#define PIN_OUT_SW_AMP			6		// スピーカーリレー制御
#define	PIN_OUT_PWR_RCH			3		// R-CH オーディオアンプ電源制御

// ボタン操作/イベント
#define SWDOWN_PRESSED          0x01
#define SWUP_PRESSED            0x02
#define SWNEXT_PRESSED          0x03
#define SWPREV_PRESSED          0x04
#define SWSELECT_PRESSED      	0x05
#define SWDESELECT_PRESSED     	0x06
#define SWPLAY_PRESSED          0x07
#define SWOPTION_PRESSED     	0x08
#define SWDOWN_LONG_PRESSED     0x09
#define SWUP_LONG_PRESSED       0x0A
#define PDOWN_PRESSED			0x0B
#define AUTO_POFF				0x0C

// 画面仕様
#define	LCD_IMAGE_BUFFER_SIZE	(8 * LCD_WIDTH)
#define MAX_SELDIR_ITEM			10		// DIR選択画面で一度に表示されるDIR数の最大数
#define MAX_SELMP3_ITEM         5       // TRACK選択画面で一度に表示される数
#define LCD_STRING_LINES		5		// LCDに表示する文字列行数(固定値)
#define LCD_ITEMS_MODE_5		5		// LCDに表示する項目数モード
#define LCD_ITEMS_MODE_10		10
#define	PLAYTIME_LEN			6		// 演奏時間表示文字数 "mmm:ss"
#define	PLAYTIME_WIDTH			(PLAYTIME_LEN*HAN_WIDTH)

// 機能
#define PLAY_NEXT_MP3           0x91
#define PLAY_PREV_MP3           0x92
#define PLAY_OTHER_DIR          0x93
#define POWEROFF                0x94
#define GOTO_MAINMENU           0x95
#define SELECT_OTHER_MP3		0x96
#define OPTION_1				0x97
#define OPTION_2				0x98
#define OPTION_3				0x99
#define DO_NOTHING				0x9A
#define	PLAY_ONCEMORE			0x9B

// madplayへ送信するコマンド(mp3_13_menuではすべてをサポートしない)
// madplay-0.15.2bのplayer.cから流用
enum {
  KEY_PAUSE    = 'p',
  KEY_STOP     = 's',
  KEY_FORWARD  = 'f',
  KEY_BACK     = 'b',
  KEY_TIME     = 't',
  KEY_QUIT     = 'q',
  KEY_INFO     = 'i',
  KEY_GAINDECR = '-',
  KEY_GAININCR = '+',
  KEY_GAINZERO = '_',
  KEY_GAININFO = '='
};

// その他定数
#define		AUTOPOFF_PLAYING_TIMEOUT		(4*60*60*1000)		// 再生中の自動シャットダウン 1hour
#define		AUTOPOFF_SELECT_TIMEOUT			(3*60*1000)			// 選択操作中の自動シャットダウン時間 3min
#define		LCDOFF_TIMEOUT					(30*1000)			// LCD表示自動オフ時間 30sec
#define		NOTIFYPIC_INTERVAL				(1*60*1000)			// RASPI->PICカラ通知時間間隔 1min
#define		PLAYTIME_INTERVAL				(5*1000)			// 演奏時間表示更新時間間隔 5sec
#define		LONGPRESS_INTERVAL				(1500)				// ボタン長押し時間 1.5sec
#define		OUTPUT_MONO			0		// モノラル出力
#define		OUTPUT_STEREO		1		// ステレオ出力
#define		DEFAULT_VOLUME		-20
#define		VOLUME_STEP			4

// イベント/制御
static volatile int	PinEvent;
static uint8_t      PonRestart;     // execute Power on Restart
static unsigned int AutoPoffTimer;
unsigned int LastEventTime;

typedef enum {PM_RANDOM_DIR, PM_RANDOM_TRACK, PM_SELECT} playmode_t;
static playmode_t   PlayMode;

typedef enum {PS_PLAY, PS_SELECT} playorselect_t;

static int		Volume;
static int 		SpRelay;
static int		StereoStatus = OUTPUT_MONO;
static int		LcdDispStatus;

// SPI
static int FdSpi;
static const char *device = "/dev/spidev0.0";
static uint8_t mode;
static uint8_t bits = 8;
static uint32_t speed = 500000;
static uint16_t delay_usecs;
static uint8_t rx[1024];

// ワーク
static uint8_t    	LcdImageBuff[LCD_IMAGE_BUFFER_SIZE]; 	// LCDフル画面イメージデータ
static uint8_t      PlayDirHist[RECLEN_HISTFILE];			// 履歴ファイルレコード
static uint16_t		DigitImage[DIGITIMAGE_LEN/2];			// 数字イメージ

// 関数宣言
void InitApp();
void PIN_IRQHandler(void);
void SetMenuItemSelected(uint8_t selecteditem, uint8_t itemsperpage);
static void systemerror(const char *s);
static void sendimage(int fd);
//static void lcddispon(int fd);
static void lcddispoff(int fd);
static void powerdown(int fd);
static void notifyactive(int fd);
static void transfer(int fd, uint8_t *tx, int tbytes);
static void DispLcdFullImage();
void DispStereoStatus();
int SelectMainMenu(int init_menu_item);
static void invert_lcd_disp_data(uint8_t lcd_disp_data[], int lineheight, int line, int x1, int x2);
int PauseAndSelectItem(int longpress);
int SelectDirectory(uint8_t *RecDirList);
int SelectIndex(uint8_t *RecDirList);
int PlayRandom(uint8_t *RecDirList);
void SetRecDirList2MenuImage(uint8_t *RecDirListImage, uint8_t no, uint8_t itemsperpage);
void SetRecDirList2IndexMenuImage(uint8_t *RecDirListImage, uint8_t no);
void SetImageAtPlaying(FILE *Fil, uint8_t *RecMp3List, uint8_t trackno, uint16_t numsndfile, uint8_t NumSelItem, uint8_t *RecDirList);
static void SetPlayTimeInImage(unsigned int play_time);
int PlayOrSelectMp3InSubdirectory(uint8_t *RecDirList, uint16_t dirno, uint16_t numsndfile, uint8_t start_trackno, playorselect_t playorselect);
int PlayOneMp3(const char *mp3file, uint16_t dirno, uint8_t trackno);
int SwNextAtPlaying(unsigned int *current_play_time);
int SwPrevAtPlaying(unsigned int *current_play_time);
void ReadDigitImage();
void ReadHistory();
void SaveHistory();
void SetRchPower();


int main(){
    int			selectedmenu;
    uint8_t     RecDirList[RECLEN_DIRLIST_FULL];
    int			ret;
    int         return_to_mainmenu;
    
	InitApp();
	ReadDigitImage();
	ReadHistory();
	SetRchPower();
	return_to_mainmenu = 1;
    AutoPoffTimer = 0;
    selectedmenu = 0;
	
	//AutoPoffTimer = millis() + AUTOPOFF_PLAYING_TIMEOUT;
	//PlayOneMp3("/snd/a.mp3", 1, 1);	// 再生テスト
	
	do {
		selectedmenu = SelectMainMenu(selectedmenu);
		switch (selectedmenu){
		case MAINMENU_SELDIR:			// ディレクトリ選択
			PlayMode = PM_SELECT;
			ret = SelectDirectory(RecDirList);
			if (ret == 0 || ret != GOTO_MAINMENU){
				if (ret == POWEROFF){
					powerdown(FdSpi);
				}
				ret = MAINMENU_POFF;
				return_to_mainmenu = 0;
			}
			break;
		case MAINMENU_SELINDEX:			// インデックスメニュー
			PlayMode = PM_SELECT;
			ret = SelectIndex(RecDirList);
			if (ret == 0 || ret != GOTO_MAINMENU){
				if (ret == POWEROFF){
					powerdown(FdSpi);
				}
				ret = MAINMENU_POFF;
				return_to_mainmenu = 0;
			}
			break;
		case MAINMENU_RANDOM:			// ランダム再生
			PlayMode = PM_RANDOM_DIR;
			PonRestart = 1;
			ret = PlayRandom(RecDirList);
			if (ret == 0 || ret != GOTO_MAINMENU){
				if (ret == POWEROFF){
					powerdown(FdSpi);
				}
				ret = MAINMENU_POFF;
				return_to_mainmenu = 0;
			}
			break;
		case MAINMENU_POFF:				// パワーオフ
			powerdown(FdSpi);
			ret = MAINMENU_POFF;
			return_to_mainmenu = 0;
			break;
		case MAINMENU_DIRSTRUCT:		// ディレクトリ構成ファイル、MP3リストファイルなど生成
			ret = MAINMENU_DIRSTRUCT;
			return_to_mainmenu = 0;
			break;
		case MAINMENU_MONOSTEREO:		// ステレオ、モノラル切り替え
			if (StereoStatus == OUTPUT_MONO){
				StereoStatus = OUTPUT_STEREO;
			} else {
				StereoStatus = OUTPUT_MONO;
			}
			SetRchPower();
			DispStereoStatus();
			break;
		}
		
	} while (return_to_mainmenu);
	
	SpRelay = 0;
	digitalWrite(PIN_OUT_SW_AMP, SpRelay);
	close(FdSpi);
	
	return ret;
}


/*
 *	初期化処理
 *	1.Wiringpi
 *	2.割り込み
 *	3.SPI
 */
void InitApp(){
	int ret = 0;
	
	LastEventTime = 0;
	wiringPiSetup();
	pinMode(0, OUTPUT);
	//
	pinMode(PIN_SWUP, INPUT);
	pullUpDnControl(PIN_SWUP, PUD_UP);
	wiringPiISR(PIN_SWUP, INT_EDGE_FALLING, PIN_IRQHandler );
	//
	pinMode(PIN_SWDOWN, INPUT);
	pullUpDnControl(PIN_SWDOWN, PUD_UP);
	wiringPiISR(PIN_SWDOWN, INT_EDGE_FALLING, PIN_IRQHandler );
	//
	pinMode(PIN_SWNEXT, INPUT);
	pullUpDnControl(PIN_SWNEXT, PUD_UP);
	wiringPiISR(PIN_SWNEXT, INT_EDGE_FALLING, PIN_IRQHandler );
	//
	pinMode(PIN_SWPREV, INPUT);
	pullUpDnControl(PIN_SWPREV, PUD_UP);
	wiringPiISR(PIN_SWPREV, INT_EDGE_FALLING, PIN_IRQHandler );
	//
	pinMode(PIN_SWSEL, INPUT);
	pullUpDnControl(PIN_SWSEL, PUD_UP);
	wiringPiISR(PIN_SWSEL, INT_EDGE_FALLING, PIN_IRQHandler);
	//
	pinMode(PIN_SWDESEL, INPUT);
	pullUpDnControl(PIN_SWDESEL, PUD_UP);
	wiringPiISR(PIN_SWDESEL, INT_EDGE_FALLING, PIN_IRQHandler);
	//
	pinMode(PIN_SWPLAY, INPUT);
	pullUpDnControl(PIN_SWPLAY, PUD_UP);
	wiringPiISR(PIN_SWPLAY, INT_EDGE_FALLING, PIN_IRQHandler);
	//
	pinMode(PIN_SWOPTION, INPUT);
	pullUpDnControl(PIN_SWOPTION, PUD_UP);
	wiringPiISR(PIN_SWOPTION, INT_EDGE_FALLING, PIN_IRQHandler);
	//
	pinMode(PIN_INP_PDOWN, INPUT);
	pullUpDnControl(PIN_INP_PDOWN, PUD_UP);
	waitForInterrupt(PIN_INP_PDOWN, -1);
	wiringPiISR(PIN_INP_PDOWN, INT_EDGE_FALLING, PIN_IRQHandler);
	//
	pinMode(PIN_OUT_SW_AMP, OUTPUT);
	SpRelay = 1;
	digitalWrite(PIN_OUT_SW_AMP, SpRelay);
	//
	pinMode(PIN_OUT_PWR_RCH, OUTPUT);
	digitalWrite(PIN_OUT_PWR_RCH, 0);
	
	// activate spi by raspi-config
	FdSpi = open(device, O_RDWR);
	if (FdSpi < 0) systemerror("SPI device error");
	
	/*
	 * spi mode
	 */
	mode = 1;	// MODE 1
	ret = ioctl(FdSpi, SPI_IOC_WR_MODE, &mode);
	if (ret == -1) systemerror("SPI SPI_IOC_WR_MODE error");
	ret = ioctl(FdSpi, SPI_IOC_RD_MODE, &mode);
	if (ret == -1) systemerror("SPI SPI_IOC_RD_MODE error");
	
	/*
	 * bits per word
	 */
	bits = 8;
	ret = ioctl(FdSpi, SPI_IOC_WR_BITS_PER_WORD, &bits);
	if (ret == -1) systemerror("SPI SPI_IOC_WR_BITS_PER_WORD error");
	ret = ioctl(FdSpi, SPI_IOC_RD_BITS_PER_WORD, &bits);
	if (ret == -1) systemerror("SPI SPI_IOC_RD_BITS_PER_WORD error");
	
	/*
	 * max speed hz
	 */
	speed = 100000;
	ret = ioctl(FdSpi, SPI_IOC_WR_MAX_SPEED_HZ, &speed);
	if (ret == -1) systemerror("SPI SPI_IOC_WR_MAX_SPEED_HZ error");
	ret = ioctl(FdSpi, SPI_IOC_RD_MAX_SPEED_HZ, &speed);
	if (ret == -1) systemerror("SPI SPI_IOC_RD_MAX_SPEED_HZ error");
	
	printf("spi mode: %d\n", mode);
	printf("bits per word: %d\n", bits);
	printf("max speed: %d Hz\n", speed);
}


// Interrupt handler
void PIN_IRQHandler(void){
	unsigned int e = millis();
	if ((e - LastEventTime) > 600){			// チャッタリング対策
		if (digitalRead(PIN_SWUP) == 0){
			PinEvent = SWUP_PRESSED;
		}
		if (digitalRead(PIN_SWDOWN) == 0){
			PinEvent = SWDOWN_PRESSED;
		}
		if (digitalRead(PIN_SWNEXT) == 0){
			PinEvent = SWNEXT_PRESSED;
		}
		if (digitalRead(PIN_SWPREV) == 0){
			PinEvent = SWPREV_PRESSED;
		}
		if (digitalRead(PIN_SWSEL) == 0){
			PinEvent = SWSELECT_PRESSED;
		}
		if (digitalRead(PIN_SWDESEL) == 0){
			PinEvent = SWDESELECT_PRESSED;
		}
		if (digitalRead(PIN_SWPLAY) == 0){
			PinEvent = SWPLAY_PRESSED;
		}
		if (digitalRead(PIN_SWOPTION) == 0){
			PinEvent = SWOPTION_PRESSED;
		}
		if (digitalRead(PIN_INP_PDOWN) == 0){
			PinEvent = PDOWN_PRESSED;
		}
		if (PinEvent != 0){
			printf("event=%d\n", PinEvent);
		}
		LastEventTime = e;
		
	}
}


// LCD全体をフル表示
static void DispLcdFullImage(){
	sendimage(FdSpi);
}


// ステレオ・モノラル選択時メッセージ
void DispStereoStatus(){
	FILE		*Fil;
	unsigned short numMenuItems;	// メインメニューの項目数
	unsigned char buf[LCD_WIDTH*2];
	
	memset(LcdImageBuff, 0, LCD_IMAGE_BUFFER_SIZE);
	
	Fil = fopen(MENUDAT13, "r");
	fread(&numMenuItems, sizeof(unsigned short), 1, Fil);
	fseek(Fil, (uint32_t)(numMenuItems) * LCD_WIDTH * sizeof(unsigned short), SEEK_CUR);
	if (StereoStatus == OUTPUT_STEREO){
		fseek(Fil, LCD_WIDTH * sizeof(unsigned short), SEEK_CUR);	// モノラルのイメージをとばす
	}
	fread(buf, sizeof(unsigned short), LCD_WIDTH, Fil);
	SetRecDirList2MenuImage(buf, 2, LCD_STRING_LINES);
   	DispLcdFullImage();
	fclose(Fil);
	
	sleep(3);	// sec
}

/*
 *	メインメニューの選択操作
 *		init_menu_item : 初期選択のメニュー項目
 */
int SelectMainMenu(int init_menu_item){
	FILE		*Fil;
	unsigned short numMenuItems;
	uint8_t     readdlflg;
    uint8_t		items = 0;
    uint16_t    idxMenu;		// 先頭に表示中のメニュー
    uint8_t		idxCurMenu;		// 現在選択中のメニュー
    uint8_t		idxCurMenuBk;	// 上記の一時的退避
    uint8_t		i;
    uint8_t		NumSelItem;				// １ページに表示する最大項目数
    int			menu_item;
    int			ret;
	unsigned char buf[LCD_WIDTH*2];
	
    readdlflg = 1;
    if (init_menu_item >= 0 && init_menu_item <= 4){
    	idxMenu = 0;
    } else {
    	idxMenu = 5;
    }
    idxCurMenu = (init_menu_item % LCD_STRING_LINES);
    NumSelItem = LCD_STRING_LINES;
	Fil = fopen(MENUDAT13, "r");
	if (Fil == NULL){
		return -1;
	}
	
	fread(&numMenuItems, sizeof(unsigned short), 1, Fil);	// メニュー項目数
	
	while (1){
    	if (readdlflg == 1){
    		items = 0;
			memset(LcdImageBuff, 0, LCD_IMAGE_BUFFER_SIZE);
			fseek(Fil, MAINMENU_ITEMS_LEN + (uint32_t)idxMenu * LCD_WIDTH * sizeof(unsigned short), SEEK_SET);
			for (i=0; i<NumSelItem && (idxMenu + i) < numMenuItems ; i++){
				if (fread(buf, sizeof(unsigned short), LCD_WIDTH, Fil) == LCD_WIDTH){
					items++;
					SetRecDirList2MenuImage(buf, i, NumSelItem);
				} else {
					// エラー
					break;
				}
			}
			SetMenuItemSelected(idxCurMenu, NumSelItem);
        	DispLcdFullImage();
    	}
    	
		ret = PauseAndSelectItem(0);
		printf("SelectMainMenu %d\n", ret);
	    readdlflg = 1;
		if (ret == SWUP_PRESSED){
			idxCurMenuBk = idxCurMenu;
			if ((idxCurMenu % LCD_STRING_LINES) == 0){
				if (idxMenu >= NumSelItem){
					idxMenu -= NumSelItem;		// 前ページへ
					idxCurMenu = LCD_STRING_LINES - 1;
				} else {
					readdlflg = 0;
				}
			} else {
				idxCurMenu--;
			}
			if (readdlflg == 0 && idxCurMenu != idxCurMenuBk){
				SetMenuItemSelected(idxCurMenuBk, NumSelItem);
				SetMenuItemSelected(idxCurMenu, NumSelItem);
	        	DispLcdFullImage();
			}
		} else if (ret == SWDOWN_PRESSED){
			idxCurMenuBk = idxCurMenu;
			if ((idxCurMenu % LCD_STRING_LINES) == (LCD_STRING_LINES - 1)){
				if ((idxMenu + NumSelItem) < numMenuItems){
					idxMenu += NumSelItem;	// 次のページへ
					idxCurMenu = 0;
				} else {
					readdlflg = 0;
				}
			} else {
				if (idxCurMenu < (items - 1)){
					idxCurMenu++;
					readdlflg = 0;
				} else {
					readdlflg = 0;
				}
			}
			if (readdlflg == 0 && idxCurMenu != idxCurMenuBk){
				SetMenuItemSelected(idxCurMenuBk, NumSelItem);
				SetMenuItemSelected(idxCurMenu, NumSelItem);
	        	DispLcdFullImage();
			}
		} else if (ret == SWSELECT_PRESSED){
			menu_item = idxMenu + idxCurMenu;
			break;
		} else if (ret == AUTO_POFF || ret == PDOWN_PRESSED){
			menu_item = MAINMENU_POFF;
			break;
		}
	}
	
	fclose(Fil);
	//printf("SelectMainMenu menu_item=%d\n", menu_item);
	return menu_item;
}


/*
 * メニュー項目の選択状態を反転する。１２DOTフォントを前提。総項目数itemsperpageは１０または５
 * 	１０のときのメニュー項目番号は
 * 		0	5
 * 		1	6
 * 		2	7
 * 		3	8
 * 		4	9
 * 	４０のときも同様に左上から下へ0,1,2,3,...の並び。
 */
void SetMenuItemSelected(uint8_t selecteditem, uint8_t itemsperpage){
	int line;
	int x1;
	int x2;

	switch (itemsperpage){
	case 10:	// 12dotフォントで5行×２列
		line = selecteditem % 5;
		if (selecteditem < 5){
			x1 = 0;
			x2 = LCD_WIDTH/2 - 1;
		} else {
			x1 = LCD_WIDTH/2;
			x2 = LCD_WIDTH - 1;
		}
		break;
	case 5:		// 12dotフォントで５行×１列
		line = selecteditem;
		x1 = 0;
		x2 = LCD_WIDTH - 1;
		break;
	case NUM_SELINDEX_ITEM:	// インデックス選択ページ
		line = selecteditem % SELINDEX_LINES;
		x1 = (selecteditem / SELINDEX_LINES) * IL_WIDTH_SELINDEX_ITEM;
		x2 = ((selecteditem / SELINDEX_LINES) +  1) * IL_WIDTH_SELINDEX_ITEM - 1;
		break;
	default:
		return;
	}

	invert_lcd_disp_data(LcdImageBuff, DL_DOTLINEHEIGHT, line, x1, x2);
}

/*
 *	反転状態のイメージデータを作成する
 *		lcd_disp_data : イメージデータ1画面すべて
 *		lineheight : １行の高さ（ドット数/ 12 or 8）
 *		line : 対象行（0, 1, 2, ...)
 *		x1, x2, 対象幅
 *	MP3_12/glcd.cから
 */
void invert_lcd_disp_data(uint8_t lcd_disp_data[], int lineheight, int line, int x1, int x2){
	int	i, x, d;
	int	page;
	int mask;
	static int start_page_12[] = {0, 1, 3, 4, 6};
	static int start_mask_12[] = {0, 2, 0, 2, 0};
	static uint8_t h12mask[] = {0xFF, 0x0F, 0xF0, 0xFF};

	if (lineheight == 12){
		page = start_page_12[line];
		mask = start_mask_12[line];
		for (i=0; i<=1; i++, page++, mask++){
			d = page * LCD_WIDTH + x1;
			for (x=x1; x<=x2; x++, d++){
				lcd_disp_data[d] ^= h12mask[mask];
			}
		}
	} else if (lineheight == 8){
		page = line;
		d = page * LCD_WIDTH + x1;
		for (x=x1; x<=x2; x++, d++){
			lcd_disp_data[d] ^= 0xFF;
		}
	}
}

// イメージデータ１画面分をSPIでPICに送信＆表示
static void sendimage(int fd){
	uint8_t		spicmd[8];
	
	spicmd[0] = CMD_IMAGEDATA;
	transfer(fd, spicmd, 1);
	
	transfer(fd, LcdImageBuff, 1024);
	
	spicmd[0] = CMD_LCDDISPON;
	transfer(fd, spicmd, 1);
	LcdDispStatus = 1;
}


// LCD表示(送付済みイメージデータを表示)
/*
static void lcddispon(int fd){
	uint8_t		spicmd[8];
	spicmd[0] = CMD_LCDDISPON;
	transfer(fd, spicmd, 1);
	LcdDispStatus = 1;
}
*/

// LCD表示を消す（バックライトのオフ）
static void lcddispoff(int fd){
	uint8_t		spicmd[8];
	spicmd[0] = CMD_LCDDISPOFF;
	transfer(fd, spicmd, 1);
	LcdDispStatus = 0;
}

// PICに停止を指示
static void powerdown(int fd){
	uint8_t		spicmd[8];
	
	spicmd[0] = CMD_POWERDOWN;
	transfer(fd, spicmd, 1);
}


// PICに本PGが実行中であることを通知
static void notifyactive(int fd){
	uint8_t		spicmd[8];
	
	spicmd[0] = CMD_ACTIVE;
	transfer(fd, spicmd, 1);
}


// SPIのデータ転送
static void transfer(int fd, uint8_t *tx, int tbytes)
{
	int ret;
	
	struct spi_ioc_transfer tr = {
		.tx_buf = (unsigned long)tx,
		.rx_buf = (unsigned long)rx,
		.len = tbytes,
		.delay_usecs = delay_usecs,
		.speed_hz = speed,
		.bits_per_word = bits,
	};

	ret = ioctl(fd, SPI_IOC_MESSAGE(1), &tr);
	if (ret < 1) systemerror("SPI SPI_IOC_MESSAGE error");

}

// システムエラー
static void systemerror(const char *s){
	perror(s);
	abort();
}


// 選択画面における入力待ち
// longpress=1:長押し対応する(UP,DOWNだけ)
int PauseAndSelectItem(int longpress){
	int		ret;
	unsigned int	start_time_ms;
	unsigned int	timeout;
	
	ret = 0;
	PinEvent = 0;
	timeout = millis() + AUTOPOFF_SELECT_TIMEOUT;
	do {
		if (PinEvent != 0
			|| (longpress == 1 && digitalRead(PIN_SWUP) == 0)
			|| (longpress == 1 && digitalRead(PIN_SWDOWN) == 0)){
			
			if (longpress == 1 && digitalRead(PIN_SWUP) == 0){
				start_time_ms = millis();
				while (digitalRead(PIN_SWUP) == 0){
					if (millis() - start_time_ms > LONGPRESS_INTERVAL){
						ret = SWUP_LONG_PRESSED;
						break;
					}
				}
				if (ret == 0){
					continue;
				}
			} else if (longpress == 1 && digitalRead(PIN_SWDOWN) == 0){
				start_time_ms = millis();
				while (digitalRead(PIN_SWDOWN) == 0){
					if (millis() - start_time_ms > LONGPRESS_INTERVAL){
						ret = SWDOWN_LONG_PRESSED;
						break;
					}
				}
				if (ret == 0){
					continue;
				}
			} else {
				ret = PinEvent;
			}
			PinEvent = 0;
			return ret;
		}
		if (millis() > timeout){
			return AUTO_POFF;
		}
	} while (1);
}


/*
 * ディレクトリの選択操作
 * 	レコード番号：ヘッダレコードを０とする。以降、ルートディレクトリを１、など。
 * 	再生中のボタン機能 [??]
 * 	NEXTボタン：同一フォルダの次のトラックへ
 * 	PLAYボタン：ディレクトリ選択へ戻る。
 * 	DESELECTボタン：同一ディレクトリのトラック選択に戻る
 */
int SelectDirectory(uint8_t *RecDirList){
	FILE		*Fil;
    uint32_t    ofsSubDirStart;
    uint16_t    rowSubDirStart;			// 現在対象のディレクトリの先頭サブディレクトリのレコード番号
    uint16_t    numSubDir;				// サブディレクトリ数
    uint16_t    idxSubDir;				// 選択ページ先頭に表示中のDIR (0,1,..., numSubDir-1)
    uint8_t     idxCurSubDirItem;   	// 現在表示ページで選択中のDIR(0,1,2,3,...,9)
    uint8_t     idxCurSubDirItemBk;		// 上記の一時的退避
    uint8_t		NumSelItem;				// １ページに表示する最大項目数
    uint8_t     idxDirDepth;
    uint16_t    stackIdxSubDir[8];
    uint8_t     stackIdxCurrentSubDirItem[8];
    uint16_t    dirno[MAX_SELDIR_ITEM];
    uint16_t    numSubDirTab[MAX_SELDIR_ITEM];
    uint16_t    numSndFile[MAX_SELDIR_ITEM];
    uint16_t    recNoFirstSubdir[MAX_SELDIR_ITEM];
    uint8_t     readdlflg;      		// 1のときDIR LIST ファイルを読む
    uint8_t		i;
    uint8_t		items = 0;
    int			ret;
	
	Fil = fopen(DIRSTRUCTUREFILE13, "r");
	if (Fil == NULL){
		return 0;
	}
	
    // Skip HEADER info
	if (fseek(Fil, RECLEN_DIRLIST_FULL, SEEK_SET) != 0){
	    fclose(Fil);
		return 0;
	}
	
    // Read ROOT info
    if (fread(RecDirList, 1, RECLEN_DIRLIST_FULL, Fil) != RECLEN_DIRLIST_FULL){
	    fclose(Fil);
    	return 0;
    }
    
	memcpy(&rowSubDirStart, RecDirList + DL_RECNO_SUBDIR_OFFSET, 2);
    ofsSubDirStart = (uint32_t)rowSubDirStart * RECLEN_DIRLIST_FULL;
    numSubDir = *(RecDirList + DL_NUM_SUBDIR_OFFSET);
    idxSubDir = 0;
    idxCurSubDirItem = 0;
    idxCurSubDirItemBk = -1;
    idxDirDepth = 0;
    readdlflg = 1;
    NumSelItem = MAX_SELDIR_ITEM;
    
    // end point of ROOT info
	if (fseek(Fil, 2*RECLEN_DIRLIST_FULL, SEEK_SET) != 0){		// ヘッダレコードとルートディレクトリレコードを読み飛ばす
		return 0;
	}
    
    while (1){
    	if (readdlflg == 1){
    		items = 0;
			memset(LcdImageBuff, 0, LCD_IMAGE_BUFFER_SIZE);
			fseek(Fil, ofsSubDirStart + (uint32_t)idxSubDir * RECLEN_DIRLIST_FULL, SEEK_SET);
			for (i=0; i<NumSelItem; i++){
				if ((idxSubDir + i) < numSubDir){
    				if (fread(RecDirList, 1, RECLEN_DIRLIST_FULL, Fil) == RECLEN_DIRLIST_FULL){
						printf(" %d [%s]%d %d \n", i, RecDirList + DL_FULLDIR_OFFSET, *(uint16_t *)(RecDirList + DL_RECNO_PARENTDIR_OFFSET), *(uint16_t *)(RecDirList + DL_RECNO_GRANDPARENTDIR_OFFSET) );
						dirno[i] = rowSubDirStart + idxSubDir + i;
                        numSubDirTab[i] = *(RecDirList + DL_NUM_SUBDIR_OFFSET);
						numSndFile[i] = *(RecDirList + DL_NUM_SNDFILE_OFFSET);
						memcpy(&recNoFirstSubdir[i], RecDirList + DL_RECNO_SUBDIR_OFFSET, 2);
						items++;
						SetRecDirList2MenuImage(RecDirList + RECLEN_DIRLIST, i, NumSelItem);
					} else {
						// エラー
					}
				} else {
					dirno[i] = 0;
					numSndFile[i] = 0;
					recNoFirstSubdir[i] = 0;
				}
			}
			SetMenuItemSelected(idxCurSubDirItem, NumSelItem);
        	DispLcdFullImage();
    	}
    	
		ret = PauseAndSelectItem(1);
	    readdlflg = 1;
		if (ret == SWNEXT_PRESSED){
			// 同一行内の項目を順次選択 ： 同一行の右の項目を選択
			idxCurSubDirItemBk = idxCurSubDirItem;
			if ((idxCurSubDirItem + LCD_STRING_LINES) < items){
				idxCurSubDirItem += LCD_STRING_LINES;
			} else {
				idxCurSubDirItem %= LCD_STRING_LINES;
			}
			if (idxCurSubDirItem != idxCurSubDirItemBk){
				SetMenuItemSelected(idxCurSubDirItemBk, NumSelItem);
				SetMenuItemSelected(idxCurSubDirItem, NumSelItem);
	        	DispLcdFullImage();
			}
			readdlflg = 0;
		} else if (ret == SWPREV_PRESSED){
			// 同一行内の項目を順次選択 ： 同一行の右の項目を選択
			idxCurSubDirItemBk = idxCurSubDirItem;
			if (idxCurSubDirItem >= LCD_STRING_LINES){
				idxCurSubDirItem -= LCD_STRING_LINES;
			} else {
				while (1){
					if ((idxCurSubDirItem + LCD_STRING_LINES) < items){
						idxCurSubDirItem += LCD_STRING_LINES;
					} else {
						break;
					}
				}
			}
			if (idxCurSubDirItem != idxCurSubDirItemBk){
				SetMenuItemSelected(idxCurSubDirItemBk, NumSelItem);
				SetMenuItemSelected(idxCurSubDirItem, NumSelItem);
	        	DispLcdFullImage();
			}
			readdlflg = 0;
		} else if (ret == SWUP_PRESSED){
			// 同一ページ内の上の行へ、または、前ページへ
			idxCurSubDirItemBk = idxCurSubDirItem;
			if ((idxCurSubDirItem % LCD_STRING_LINES) == 0){
				if (idxSubDir >= NumSelItem){
					idxSubDir -= NumSelItem;	// 前ページへ
					if (idxCurSubDirItem < LCD_STRING_LINES){
	               		idxCurSubDirItem = LCD_STRING_LINES - 1;
	               	} else {
	               		idxCurSubDirItem = NumSelItem - 1;
	               	}
				} else {
					readdlflg = 0;
				}
			} else {
				idxCurSubDirItem--;
				readdlflg = 0;
			}
			if (readdlflg == 0 && idxCurSubDirItem != idxCurSubDirItemBk){
				SetMenuItemSelected(idxCurSubDirItemBk, NumSelItem);
				SetMenuItemSelected(idxCurSubDirItem, NumSelItem);
	        	DispLcdFullImage();
			}
		} else if (ret == SWUP_LONG_PRESSED){
			if (idxSubDir >= NumSelItem){
				idxSubDir -= NumSelItem;	// 前ページへ
				if (idxCurSubDirItem < LCD_STRING_LINES){
               		idxCurSubDirItem = LCD_STRING_LINES - 1;
               	} else {
               		idxCurSubDirItem = NumSelItem - 1;
               	}
			} else {
				readdlflg = 0;
			}
		} else if (ret == SWDOWN_PRESSED){
			// 同一ページ内の下の行へ、または、次ページへ
			idxCurSubDirItemBk = idxCurSubDirItem;
			if ((idxCurSubDirItem % LCD_STRING_LINES) == (LCD_STRING_LINES - 1)){
				if ((idxSubDir + NumSelItem) < numSubDir){
					idxSubDir += NumSelItem;	// 次のページへ
					idxCurSubDirItem = 0;
				} else {
					readdlflg = 0;
				}
			} else {
				if (idxCurSubDirItem < (items - 1)){
					idxCurSubDirItem++;
					readdlflg = 0;
				} else {
					readdlflg = 0;
				}
			}
			if (readdlflg == FALSE && idxCurSubDirItem != idxCurSubDirItemBk){
				SetMenuItemSelected(idxCurSubDirItemBk, NumSelItem);
				SetMenuItemSelected(idxCurSubDirItem, NumSelItem);
	        	DispLcdFullImage();
			}
		} else if (ret == SWDOWN_LONG_PRESSED){
			if ((idxSubDir + NumSelItem) < numSubDir){
				idxSubDir += NumSelItem;	// 次のページへ
				idxCurSubDirItem = 0;
			} else {
				readdlflg = 0;
			}
		} else if (ret == SWPLAY_PRESSED){		// PLAY
			if (numSndFile[idxCurSubDirItem] != 0){      // Have a MP3 file ?
				AutoPoffTimer = millis() + AUTOPOFF_PLAYING_TIMEOUT;
				fseek(Fil, ofsSubDirStart + (uint32_t)(idxSubDir + idxCurSubDirItem) * RECLEN_DIRLIST_FULL, SEEK_SET);
				fread(RecDirList, 1, RECLEN_DIRLIST_FULL, Fil);
				ret = PlayOrSelectMp3InSubdirectory(RecDirList, dirno[idxCurSubDirItem], numSndFile[idxCurSubDirItem], 0, PS_PLAY);
				if (ret == POWEROFF){
					break;
				} else if (ret == GOTO_MAINMENU){
					break;
				}
			}
		} else if (ret == SWSELECT_PRESSED){  // 下位DIRへ (SKIP forward directory)
			if (recNoFirstSubdir[idxCurSubDirItem] > 1){		// 下位DIRが存在する
				rowSubDirStart = recNoFirstSubdir[idxCurSubDirItem];
				ofsSubDirStart = (uint32_t)rowSubDirStart * RECLEN_DIRLIST_FULL;
				numSubDir = numSubDirTab[idxCurSubDirItem];		// *(RecDirList + DL_NUM_SUBDIR);
				// push current idxSubDir to stack
				stackIdxSubDir[idxDirDepth] = idxSubDir;
				stackIdxCurrentSubDirItem[idxDirDepth] = idxCurSubDirItem;
				idxDirDepth++;
				idxSubDir = 0;
				idxCurSubDirItem = 0;
				printf("SWSELECT_PRESSED rowSubDirStart=%d \n", rowSubDirStart);
		        //NumSelItem = MAX_SELDIR_ITEM;		//初期値に戻さない
			} else if (numSndFile[idxCurSubDirItem] != 0){      // Have a MP3 file ?
				fseek(Fil, ofsSubDirStart + (uint32_t)(idxSubDir + idxCurSubDirItem) * RECLEN_DIRLIST_FULL, SEEK_SET);
				fread(RecDirList, 1, RECLEN_DIRLIST_FULL, Fil);
				ret = PlayOrSelectMp3InSubdirectory(RecDirList, dirno[idxCurSubDirItem], numSndFile[idxCurSubDirItem], 0, PS_SELECT);
				if (ret == POWEROFF){
					break;
				} else if (ret == GOTO_MAINMENU){
					break;
				}
			}
		} else if (ret == SWDESELECT_PRESSED){  // 上位DIRへ (GOTO Mainmenu)
			uint16_t    r;
			uint16_t    r2;

			memcpy(&r,  RecDirList + DL_RECNO_GRANDPARENTDIR_OFFSET, 2);	// rec no of grand parent
			memcpy(&r2, RecDirList + DL_RECNO_PARENTDIR_OFFSET, 2);
			// 親々ディレクトリのレコード番号と親ディレクトリのレコード番号が一致しているときは、トップレベルのDIR階層
			if (r == r2){
				ret = GOTO_MAINMENU;
				break;
			} else {
				// Go to grand parent record
				if (fseek(Fil, (uint32_t)r * RECLEN_DIRLIST_FULL, SEEK_SET) != 0) return 0;
				// Read grand parent record
				if (fread(RecDirList, 1, RECLEN_DIRLIST_FULL, Fil) != RECLEN_DIRLIST_FULL) return 0;
				// the first "uncle叔父" record
				memcpy(&r, RecDirList + DL_RECNO_SUBDIR_OFFSET, 2);
				if (r > 1){
					rowSubDirStart = r;
					ofsSubDirStart = (uint32_t)rowSubDirStart * RECLEN_DIRLIST_FULL;
					numSubDir = *(RecDirList + DL_NUM_SUBDIR_OFFSET);
					// pop idxSubDir from stack -> This is parent
					idxDirDepth--;
					idxSubDir = stackIdxSubDir[idxDirDepth];
					idxCurSubDirItem = stackIdxCurrentSubDirItem[idxDirDepth];
					if (idxSubDir >= numSubDir) idxSubDir = 0;
			        //NumSelItem = MAX_SELDIR_ITEM;		// 初期値に戻さない
				} else {
					ret = GOTO_MAINMENU;
					break;
				}
			}
		} else if (ret == SWOPTION_PRESSED){	// オプション＝表示項目数の変更
			if (NumSelItem == MAX_SELDIR_ITEM){
				NumSelItem = MAX_SELDIR_ITEM/2;
				if (idxCurSubDirItem >= LCD_STRING_LINES){
					idxCurSubDirItem -= LCD_STRING_LINES;
				}
			} else {
				NumSelItem = MAX_SELDIR_ITEM;
			}
		} else if (ret == AUTO_POFF || ret == PDOWN_PRESSED){
			ret = POWEROFF;
			break;
	    }
    }
    
    fclose(Fil);
    
    return ret;
}



/*
 *	インデックスの選択処理
 */
int SelectIndex(uint8_t *RecDirList){
	FILE		*Fil;
    uint32_t    ofsSubDirStart;
    uint16_t    rowSubDirStart;
    uint16_t    numSubDir;
    uint16_t    idxSubDir;				// 選択ページ先頭に表示中のDIR (0,1,..., numSubDir-1)
    uint8_t     idxCurSubDirItem;   	// 現在表示ページで選択中のDIR(0,1,2,3)
    uint8_t     idxCurSubDirItemBk;		// 上記の一時的退避
    uint8_t		NumSelItem;				// １ページに表示する項目数

    int        	ret;
    uint8_t     idxDirDepth;
    uint16_t    stackIdxSubDir[8];
    uint8_t     stackIdxCurrentSubDirItem[8];
    uint16_t    dirno[NUM_SELINDEX_ITEM];			// ディレクトリ構成ファイルのレコード番号
    uint16_t    numSubDirTab[NUM_SELINDEX_ITEM];
    uint16_t    numSndFile[NUM_SELINDEX_ITEM];
    uint16_t    recNoFirstSubdir[NUM_SELINDEX_ITEM];
    uint8_t     readdlflg;      // 1のときDIR LIST ファイルを読む
    uint8_t		i;
    uint8_t		items = 0;
    int			longpress;		// 長押し対応
    
	Fil = fopen(INDEXFILE, "r");
	if (Fil == NULL){
		return 0;
	}
	
    // Skip HEADER info
	if (fseek(Fil, RECLEN_DIRLIST_FULL, SEEK_SET) != 0){
	    fclose(Fil);
		return 0;
	}
	
    // Read ROOT info
    if (fread(RecDirList, 1, RECLEN_DIRLIST_FULL, Fil) != RECLEN_DIRLIST_FULL){
	    fclose(Fil);
    	return 0;
    }
    
	memcpy(&rowSubDirStart, RecDirList + DL_RECNO_SUBDIR_OFFSET, 2);				// ルートレコードからサブディレクトリの先頭レコード番号を取得
    ofsSubDirStart = (uint32_t)rowSubDirStart * RECLEN_DIRLIST_FULL;
    numSubDir = *(RecDirList + DL_NUM_SUBDIR_OFFSET);							// ルートディレクトリからサブディレクトリ数（インデックス数）を取得
	printf("%d %d\n", ofsSubDirStart, numSubDir);
    idxSubDir = 0;
    idxCurSubDirItem = 0;
    idxCurSubDirItemBk = -1;
    idxDirDepth = 0;
    readdlflg = 1;
    NumSelItem = NUM_SELINDEX_ITEM;		// インデックス選択の項目数（4行 X １行あたり１０項目）
    longpress = 0;
    
    // end point of ROOT info
	if (fseek(Fil, 2*RECLEN_DIRLIST_FULL, SEEK_SET) != 0){		// ヘッダレコードとルートディレクトリレコードを読み飛ばす
		return 0;
	}
    
    while (1){
    	if (readdlflg == 1){
    		items = 0;
			memset(LcdImageBuff, 0, LCD_IMAGE_BUFFER_SIZE);
			fseek(Fil, ofsSubDirStart + (uint32_t)idxSubDir * RECLEN_DIRLIST_FULL, SEEK_SET);
			for (i=0; i<NumSelItem; i++){
				if ((idxSubDir + i) < numSubDir){
    				if (fread(RecDirList, 1, RECLEN_DIRLIST_FULL, Fil) == RECLEN_DIRLIST_FULL){
						//dirno[i] = rowSubDirStart + idxSubDir - 1;		// インデックスファイルのレコード番号
						memcpy(&dirno[i], RecDirList + IL_DLRECCNO_OFFSET, IL_DLRECNO_LEN);
                        numSubDirTab[i] = *(RecDirList + DL_NUM_SUBDIR_OFFSET);
						numSndFile[i] = *(RecDirList + DL_NUM_SNDFILE_OFFSET);
						memcpy(&recNoFirstSubdir[i], RecDirList + DL_RECNO_SUBDIR_OFFSET, 2);
						items++;
						if (idxDirDepth == 0){	// インデックスのトップページ(A,B,C,...フォルダの頭文字順)
							SetRecDirList2IndexMenuImage(RecDirList + RECLEN_DIRLIST, i);
						} else {	// インデックスのサブページ（各アルバム・作品などMP3ファイルを持つフォルダ)
							SetRecDirList2MenuImage(RecDirList + RECLEN_DIRLIST, i, NumSelItem);
						}
					} else {
						// エラー
					}
				} else {
					dirno[i] = 0;
					numSndFile[i] = 0;
					recNoFirstSubdir[i] = 0;
				}
			}
			SetMenuItemSelected(idxCurSubDirItem, NumSelItem);
        	DispLcdFullImage();
    	}
    	
		ret = PauseAndSelectItem(longpress);
	    readdlflg = 1;
		if (ret == SWNEXT_PRESSED){
			// 同一行内の項目を順次選択 ： 同一行の右の項目を選択
			idxCurSubDirItemBk = idxCurSubDirItem;
			if ((idxCurSubDirItem + SELINDEX_LINES) < items){
				idxCurSubDirItem += SELINDEX_LINES;
			} else {
				idxCurSubDirItem %= SELINDEX_LINES;
			}
			if (idxCurSubDirItem != idxCurSubDirItemBk){
				SetMenuItemSelected(idxCurSubDirItemBk, NumSelItem);
				SetMenuItemSelected(idxCurSubDirItem, NumSelItem);
	        	DispLcdFullImage();
			}
			readdlflg = 0;
		} else if (ret == SWPREV_PRESSED){
			// 同一行内の項目を順次選択 ： 同一行の右の項目を選択
			idxCurSubDirItemBk = idxCurSubDirItem;
			if (idxCurSubDirItem >= SELINDEX_LINES){
				idxCurSubDirItem -= SELINDEX_LINES;
			} else {
				while (1){
					if ((idxCurSubDirItem + SELINDEX_LINES) < items){
						idxCurSubDirItem += SELINDEX_LINES;
					} else {
						break;
					}
				}
			}
			if (idxCurSubDirItem != idxCurSubDirItemBk){
				SetMenuItemSelected(idxCurSubDirItemBk, NumSelItem);
				SetMenuItemSelected(idxCurSubDirItem, NumSelItem);
	        	DispLcdFullImage();
			}
			readdlflg = 0;
		} else if (ret == SWUP_PRESSED){
			// 同一ページ内の上の行へ、または、前ページへ
			idxCurSubDirItemBk = idxCurSubDirItem;
			if ((idxCurSubDirItem % SELINDEX_LINES) == 0){
				if (idxSubDir >= NumSelItem){
					idxSubDir -= NumSelItem;	// 前ページへ
					idxCurSubDirItem = 0;
				} else {
					readdlflg = 0;
				}
			} else {
				idxCurSubDirItem--;
				readdlflg = 0;
			}
			if (readdlflg == 0 && idxCurSubDirItem != idxCurSubDirItemBk){
				SetMenuItemSelected(idxCurSubDirItemBk, NumSelItem);
				SetMenuItemSelected(idxCurSubDirItem, NumSelItem);
	        	DispLcdFullImage();
			}
		} else if (ret == SWUP_LONG_PRESSED){
			if (idxSubDir >= NumSelItem){
				idxSubDir -= NumSelItem;	// 前ページへ
				idxCurSubDirItem = 0;
			} else {
				readdlflg = 0;
			}
		} else if (ret == SWDOWN_PRESSED){
			// 同一ページ内の下の行へ、または、次ページへ
			idxCurSubDirItemBk = idxCurSubDirItem;
			if ((idxCurSubDirItem % SELINDEX_LINES) == (SELINDEX_LINES - 1)){
				if ((idxSubDir + NumSelItem) < numSubDir){
					idxSubDir += NumSelItem;	// 次のページへ
					idxCurSubDirItem = 0;
				} else {
					readdlflg = 0;
				}
			} else {
				if (idxCurSubDirItem < (items - 1)){
					idxCurSubDirItem++;
					readdlflg = 0;
				} else {
					readdlflg = 0;
				}
			}
			if (readdlflg == FALSE && idxCurSubDirItem != idxCurSubDirItemBk){
				SetMenuItemSelected(idxCurSubDirItemBk, NumSelItem);
				SetMenuItemSelected(idxCurSubDirItem, NumSelItem);
	        	DispLcdFullImage();
			}
		} else if (ret == SWDOWN_LONG_PRESSED){
			if ((idxSubDir + NumSelItem) < numSubDir){
				idxSubDir += NumSelItem;	// 次のページへ
				idxCurSubDirItem = 0;
			} else {
				readdlflg = 0;
			}
		} else if (ret == SWPLAY_PRESSED){		// PLAY
			if (numSndFile[idxCurSubDirItem] != 0){      // Have a MP3 file ?
				AutoPoffTimer = millis() + AUTOPOFF_PLAYING_TIMEOUT;
				fseek(Fil, ofsSubDirStart + (uint32_t)(idxSubDir + idxCurSubDirItem) * RECLEN_DIRLIST_FULL, SEEK_SET);
				fread(RecDirList, 1, RECLEN_DIRLIST_FULL, Fil);
				ret = PlayOrSelectMp3InSubdirectory(RecDirList, dirno[idxCurSubDirItem], numSndFile[idxCurSubDirItem], 0, PS_PLAY);
				if (ret == POWEROFF){
					break;
				} else if (ret == GOTO_MAINMENU){
					break;
				}
			}
		} else if (ret == SWSELECT_PRESSED){  // 下位DIRへ (SKIP forward directory)
			if (recNoFirstSubdir[idxCurSubDirItem] > 1){		// 下位DIRが存在する
				rowSubDirStart = recNoFirstSubdir[idxCurSubDirItem];
				ofsSubDirStart = (uint32_t)rowSubDirStart * RECLEN_DIRLIST_FULL;
				numSubDir = numSubDirTab[idxCurSubDirItem];		// *(RecDirList + DL_NUM_SUBDIR);
				// push current idxSubDir to stack
				stackIdxSubDir[idxDirDepth] = idxSubDir;
				stackIdxCurrentSubDirItem[idxDirDepth] = idxCurSubDirItem;
				idxDirDepth++;
				idxSubDir = 0;
				idxCurSubDirItem = 0;
		        NumSelItem = MAX_SELDIR_ITEM;
		        longpress = 1;
			} else if (numSndFile[idxCurSubDirItem] != 0){      // Have a MP3 file ?
				fseek(Fil, ofsSubDirStart + (uint32_t)(idxSubDir + idxCurSubDirItem) * RECLEN_DIRLIST_FULL, SEEK_SET);
				fread(RecDirList, 1, RECLEN_DIRLIST_FULL, Fil);
				ret = PlayOrSelectMp3InSubdirectory(RecDirList, dirno[idxCurSubDirItem], numSndFile[idxCurSubDirItem], 0, PS_SELECT);
				if (ret == POWEROFF){
					break;
				} else if (ret == GOTO_MAINMENU){
					break;
				}
			}
		} else if (ret == SWDESELECT_PRESSED){  // 上位DIRへ (GOTO Mainmenu)
			uint16_t    r;
			uint16_t    r2;

			memcpy(&r,  RecDirList + DL_RECNO_GRANDPARENTDIR_OFFSET, 2);	// rec no of grand parent
			memcpy(&r2, RecDirList + DL_RECNO_PARENTDIR_OFFSET, 2);
			printf("r=%d r2=%d\n", r, r2);
			// 親々ディレクトリのレコード番号と親ディレクトリのレコード番号が一致しているときは、トップレベルのDIR階層
			if (r == r2){
				ret = GOTO_MAINMENU;
				break;
			} else {
				// Go to grand parent record
				if (fseek(Fil, (uint32_t)r * RECLEN_DIRLIST_FULL, SEEK_SET) != 0) return 0;
				// Read grand parent record
				if (fread(RecDirList, 1, RECLEN_DIRLIST_FULL, Fil) != RECLEN_DIRLIST_FULL) return 0;
				// the first "uncle叔父" record
				memcpy(&r, RecDirList + DL_RECNO_SUBDIR_OFFSET, 2);
				printf(" %d [%s]\n", r, RecDirList + DL_FULLDIR_OFFSET);
				if (r > 1){
					rowSubDirStart = r;
					ofsSubDirStart = (uint32_t)rowSubDirStart * RECLEN_DIRLIST_FULL;
					numSubDir = *(RecDirList + DL_NUM_SUBDIR_OFFSET);
					// pop idxSubDir from stack -> This is parent
					idxDirDepth--;
					idxSubDir = stackIdxSubDir[idxDirDepth];
					idxCurSubDirItem = stackIdxCurrentSubDirItem[idxDirDepth];
					if (idxSubDir >= numSubDir) idxSubDir = 0;
			        NumSelItem = NUM_SELINDEX_ITEM;
			        longpress = 0;
				} else {
					ret = GOTO_MAINMENU;
					break;
				}
			}
			readdlflg = 1;
		} else if (ret == SWOPTION_PRESSED){	// オプション＝表示項目数の変更
			if (NumSelItem == MAX_SELDIR_ITEM){
				NumSelItem = MAX_SELDIR_ITEM/2;
				if (idxCurSubDirItem >= LCD_STRING_LINES){
					idxCurSubDirItem -= LCD_STRING_LINES;
				}
			} else if (NumSelItem == (MAX_SELDIR_ITEM / 2)){
				NumSelItem = MAX_SELDIR_ITEM;
			} else {
				readdlflg = 0;
			}
		} else if (ret == AUTO_POFF || ret == PDOWN_PRESSED){
			ret = POWEROFF;
			break;
	    }
    }
    
    fclose(Fil);
    
    return ret;
}

/*
 * ディレクトリリストファイル/MP3リストファイル のレコードからディレクトリ選択メニューイメージを作成する。
 * 	no : メニュー項目の序数 (0 - 9)
 * 	itemsperpage : 一度に表示する項目数(10 or 8 or 5 or 4)
 */
void SetRecDirList2MenuImage(uint8_t *RecDirListImage, uint8_t no, uint8_t itemsperpage){
	static uint8_t	no2page_10[10] = {0, 1, 3, 4, 6, 0, 1, 3, 4, 6};	// itemsperpage=10/5のとき、各noのイメージをセットするページ(no=0は0と1、no=1は1と2、no=2は3と4)
	static uint8_t	halfflg_10[10] = {0, 1, 0, 1, 0, 0, 1, 0, 1, 0};	// 0はnoの項目が１ページすべてと次のページの上半に渡ってセットされる。1はnoの項目がページの下半と次のページすべてに渡ってセットされる
	static uint8_t	no2page_8[8]   = {0, 1, 3, 4, 0, 1, 3, 4};			// itemsperpage=8/4のとき、同上（各noは２つのページを使う。）
	static uint8_t	halfflg_8[8]   = {0, 1, 0, 1, 0, 1, 0, 1};			//

	int		x, page;	// pageはLCDのドキュメントで言うところのページ（縦８ビット、横６４ビットLCD上エリア）
	int		x_start, width;
	uint8_t	halfflg;
	unsigned short *image = (unsigned short *)RecDirListImage;

	if (itemsperpage == 10 || itemsperpage == 5){
		page = no2page_10[no];
		if (no >= 5){	// LCDの右半分エリアに出力
			width = LCD_WIDTH / 2 - 1;
			x_start = page *LCD_WIDTH + LCD_WIDTH / 2 + 1;
		} else {		// LCDの左半分エリアに出力するか、左から右まで使って出力する
			x_start = page *LCD_WIDTH;
			if (itemsperpage == 10){	// LCDの左半分エリアに出力する
				width = LCD_WIDTH / 2 - 1;
			} else {	// 左から右まで使って出力する
				width = LCD_WIDTH;
			}
		}
		halfflg = halfflg_10[no];
	} else {	// 8 or 4 (MP3_13ではこれらの値を使わない予定 )
		page = no2page_8[no];
		if (no >= 4){	// LCDの右半分エリアに出力
			x_start = page *LCD_WIDTH + LCD_WIDTH / 2 + 1;
			width = LCD_WIDTH / 2 - 1;
		} else {		// LCDの左半分エリアに出力するか、左から右まで使って出力する
			x_start = page *LCD_WIDTH;
			if (itemsperpage == 8){	// LCDの左半分エリアに出力する
				width = LCD_WIDTH / 2 - 1;
			} else {	// 左から右まで使って出力する
				width = LCD_WIDTH;
			}
		}
		halfflg = halfflg_8[no];
	}

	if (halfflg == 0){
		for (x=0; x<width; x++){
			LcdImageBuff[x_start + x] = *(image + x) & 0xFF;
			LcdImageBuff[x_start + LCD_WIDTH + x] = (*(image + x) >> 8) & 0x0F;
		}
	} else {
		for (x=0; x<width; x++){
			LcdImageBuff[x_start + x] |= ((*(image + x) << 4) & 0xF0);
			LcdImageBuff[x_start + LCD_WIDTH + x] = (*(image + x) >> 4) & 0xFF;
		}
	}
}


/*
 * インデックスファイル/MP3リストファイル のレコードからインデックス選択メニューイメージを作成する。
 *  RecDirList : インデックスファイルのレコード
 * 	no : メニュー項目の序数 (0～、左上から下へ0,1,2,3,... )
 */
void SetRecDirList2IndexMenuImage(uint8_t *RecDirListImage, uint8_t no){
	static uint8_t	no2page[5]   = {0, 1, 3, 4, 6};			// インデックスの１～４行目の各行の表示開始ページ(LCDページ)
	static uint8_t	halfflg[5]   = {0, 1, 0, 1, 0};			// 0:1ページ8ドット+(1/2)ページ4ドットで描画、1:(1/2)ページ４ドット+１ページ8ドットで描画

	int		x, page;
	int		x_start;
	unsigned short *image = (unsigned short *)RecDirListImage;

	page = no2page[no % SELINDEX_LINES];

	x_start = (no / SELINDEX_LINES) * IL_WIDTH_SELINDEX_ITEM;

	if (halfflg[no % SELINDEX_LINES] == 0){
		for (x=0; x<IL_WIDTH_SELINDEX_ITEM; x++){
			LcdImageBuff[page * LCD_WIDTH + x_start + x] = (image[x] & 0xFF);		// RecDirList[x];
			LcdImageBuff[(page + 1) * LCD_WIDTH + x_start + x] = ((image[x] >> 8) & 0x0F); 	//(RecDirList[RECLEN_DIRNAMEIMG_THIRD + x] & 0x0F);
		}
	} else {
		for (x=0; x<IL_WIDTH_SELINDEX_ITEM; x++){
			LcdImageBuff[page * LCD_WIDTH + x_start + x] |= ((image[x] & 0x0F) << 4);	// ((RecDirList[x] & 0x0F) << 4);
			LcdImageBuff[(page + 1) * LCD_WIDTH + x_start + x] = ((image[x] >> 4) & 0xFF);	// (((RecDirList[x] & 0xF0) >> 4) | ((RecDirList[RECLEN_DIRNAMEIMG_THIRD + x] & 0x0F) << 4));
		}
	}
}


// 現在のサブディレクトリからMP3を選択する。または再生開始する。
int PlayOrSelectMp3InSubdirectory(uint8_t *RecDirList, uint16_t dirno, uint16_t numsndfile, uint8_t start_trackno, playorselect_t playorselect){
    int		    ret;
    FILE		*filML;
    uint8_t     RecMp3List[RECLEN_MP3LIST_FULL];   // for MP3 list file
    uint8_t     trackno;		// 再生中の曲番 (0,1,2,...,numsndfile-1)
    uint8_t     idxtrack;       // 選択操作中のページ先頭の曲番 (0,1,2,...,numsndfile-1)
    uint8_t		NumSelItem;		// １ページに表示する項目数
    int		 	volinit = 0;
    uint16_t    dirlen;
    uint8_t     execplayflg = 0;
    uint8_t     readdlflg;      // 1のときMP3 LIST ファイルを読む
    uint8_t		i;
    uint8_t		items = 0;
    uint16_t    idxCurTitleItem;
    uint16_t    idxCurTitleItemBk;

    memcpy(&dirlen, RecDirList + DL_LENFULLDIR_OFFSET, DL_LENFULLDIR_LEN);
    
    // Volume file
    *(RecDirList + DL_FULLDIR_OFFSET + dirlen) = '/';
    strcpy((char *)(RecDirList + DL_FULLDIR_OFFSET + dirlen + 1), VOLUMEFILE);
    filML = fopen((char *)(RecDirList + DL_FULLDIR_OFFSET), "r");
    if (filML != NULL){
    	if (fread(&Volume, sizeof(int), 1, filML) != 1){
		    Volume = DEFAULT_VOLUME;
    	}
    	fclose(filML);
    } else {
	    Volume = DEFAULT_VOLUME;
    }
	volinit = Volume;
	
    // MP3 file list
    *(RecDirList + DL_FULLDIR_OFFSET + dirlen) = '/';
    strcpy((char *)(RecDirList + DL_FULLDIR_OFFSET + dirlen + 1), SNDFILELISTFILE);
    filML = fopen((char *)(RecDirList + DL_FULLDIR_OFFSET), "r");
    if (filML == NULL) return 0;
    
    if (start_trackno >= numsndfile) start_trackno = 0;
    trackno = start_trackno;
    idxtrack = 0;
    idxCurTitleItem = 0;
    readdlflg = 1;
    NumSelItem = MAX_SELMP3_ITEM;
    
    while (1){
        if (PlayMode == PM_SELECT && playorselect == PS_SELECT){	// 選択操作を行う
        	if (readdlflg == 1){		//
        		// 選択肢を表示
        		items = 0;
    			memset(LcdImageBuff, 0, LCD_IMAGE_BUFFER_SIZE);
    			fseek(filML, (uint32_t)idxtrack * RECLEN_MP3LIST_FULL, SEEK_SET);
        		for (i=0; i<NumSelItem; i++){
        			if ((idxtrack + i) < numsndfile){
        				if (fread(RecMp3List, 1, RECLEN_MP3LIST_FULL, filML) == RECLEN_MP3LIST_FULL){
    						items++;
  							SetRecDirList2MenuImage(RecMp3List + RECLEN_MP3LIST, i, NumSelItem);
    					} else {
    						// エラー
    					}
        			} else {
    					//memset(&(OledImageWork[i*SELMP3_ITEM_WIDTH]), 0, SELMP3_ITEM_WIDTH);
        			}
        		}
    			SetMenuItemSelected(idxCurTitleItem, NumSelItem);
            	DispLcdFullImage();
        	}
        	
            ret = PauseAndSelectItem(1);
    	    readdlflg = 1;
            if (ret == SWNEXT_PRESSED){          // NEXT mp3 file
    			readdlflg = 0;
                continue;		// 選択操作続行
            } else if (ret == SWPREV_PRESSED){	// PREV mp3 file
    			readdlflg = 0;
                continue;
            } else if (ret == SWUP_PRESSED){
            	idxCurTitleItemBk = idxCurTitleItem;
            	if (idxCurTitleItem == 0){
        			//idxCurTitleItem = NumSelItem - 1;
                	if (idxtrack >= NumSelItem){
                		idxtrack -= NumSelItem;
                		idxCurTitleItem = NumSelItem - 1;
                	} else {
            			readdlflg = 0;
                	}
            	} else {
            		idxCurTitleItem--;
        			readdlflg = 0;
            	}
    			if (readdlflg == 0 && idxCurTitleItem != idxCurTitleItemBk){
    				SetMenuItemSelected(idxCurTitleItemBk, NumSelItem);
    				SetMenuItemSelected(idxCurTitleItem, NumSelItem);
    	        	DispLcdFullImage();
    			}
                continue;
            } else if (ret == SWUP_LONG_PRESSED){
            	if (idxtrack >= NumSelItem){
            		idxtrack -= NumSelItem;
            		idxCurTitleItem = NumSelItem - 1;
            	} else {
        			readdlflg = 0;
            	}
                continue;
            } else if (ret == SWDOWN_PRESSED){
            	idxCurTitleItemBk = idxCurTitleItem;
				if (idxCurTitleItem < (NumSelItem - 1)){
					if ((idxtrack + idxCurTitleItem) < (numsndfile - 1)){
						idxCurTitleItem++;
						readdlflg = 0;
					} else {
	        			readdlflg = 0;
					}
				} else {
	            	if ((idxtrack + NumSelItem) < numsndfile){
	            		idxtrack += NumSelItem;
	            		idxCurTitleItem = 0;
	            	} else {
	        			readdlflg = 0;
	            	}
				}
    			if (readdlflg == FALSE && idxCurTitleItem != idxCurTitleItemBk){
    				SetMenuItemSelected(idxCurTitleItemBk, NumSelItem);
    				SetMenuItemSelected(idxCurTitleItem, NumSelItem);
    	        	DispLcdFullImage();
    			}
            	continue;
            } else if (ret == SWDOWN_LONG_PRESSED){
            	if ((idxtrack + NumSelItem) < numsndfile){
            		idxtrack += NumSelItem;
            		idxCurTitleItem = 0;
            	} else {
        			readdlflg = 0;
            	}
            	continue;
            } else if (ret == SWPLAY_PRESSED){
            	printf("SWPLAY_PRESSED\n");
                playorselect = PS_PLAY;		// 再生開始
                trackno = idxtrack + idxCurTitleItem;
        		NumSelItem = MAX_SELMP3_ITEM;
				AutoPoffTimer = millis() + AUTOPOFF_PLAYING_TIMEOUT;
                continue;		// いったんループの先頭に戻る
            } else if (ret == SWDESELECT_PRESSED){
            	ret = PLAY_OTHER_DIR;
        		NumSelItem = MAX_SELMP3_ITEM;
                break;			// 現ディレクトリを抜ける。
			} else if (ret == AUTO_POFF || ret == PDOWN_PRESSED){
				ret = POWEROFF;
				break;
            } else {
    			readdlflg = 0;
                continue;
            }
        } else {	// 再生開始
        	printf(" trackno=%d numsndfile=%d, NumSelItem=%d\n", trackno, numsndfile, NumSelItem);
        	SetImageAtPlaying(filML, RecMp3List, trackno, numsndfile, NumSelItem, RecDirList);
        	//SetPlayTimeInImage(0);
        	//DispLcdFullImage();		// ここでは表示しない。再開時間も含めて後で表示
        	if (fseek(filML, (uint32_t)(trackno)*(RECLEN_MP3LIST_FULL), SEEK_SET) != 0) break;
        	if (fread(RecMp3List, 1, RECLEN_MP3LIST_FULL, filML) != RECLEN_MP3LIST_FULL) break;
        }
        
        // Mp3 file full path
        *(RecDirList + DL_FULLDIR_OFFSET + dirlen) = '/';
        strcpy((char *)(RecDirList + DL_FULLDIR_OFFSET + dirlen + 1), (char *)(RecMp3List + ML_FILENAME_OFFSET));
        // Play a MP3 file
        ret = PlayOneMp3((char *)(RecDirList + DL_FULLDIR_OFFSET), dirno, trackno);
        if (ret == PLAY_OTHER_DIR){
            break;
        } else if (ret == POWEROFF){
            break;
        } else if (ret == GOTO_MAINMENU){
            break;
        } else if (ret == SELECT_OTHER_MP3){
            if (PlayMode == PM_RANDOM_DIR){
            	ret = GOTO_MAINMENU;
            	break;
            } else {
                playorselect = PS_SELECT;
            }
        } else if (ret == PLAY_PREV_MP3){
            if (trackno > 0) trackno--;
        } else {
            execplayflg = 1;
            trackno++;
            if (trackno == numsndfile) break;
        }
    }
    
    fclose(filML);
    
    // VOLUMEFILE保存
    if (execplayflg == 1 && Volume != volinit){     // at least one MP3 was played
	    *(RecDirList + DL_FULLDIR_OFFSET + dirlen) = '/';
	    strcpy((char *)(RecDirList + DL_FULLDIR_OFFSET + dirlen + 1), VOLUMEFILE);
	    filML = fopen((char *)(RecDirList + DL_FULLDIR_OFFSET), "w");
	    if (filML != NULL){
	    	fwrite(&Volume, sizeof(int), 1, filML);
	    	fclose(filML);
	    }
    }
    
    return ret;
}


/*
 * ランダム再生
 * 	再生中のボタン機能
 * 	NEXTボタン：同一フォルダの次のトラックへ
 * 	PLAYボタン：ランダム再生の次のディレクトリへ
 * 	DESELECTボタン：メインメニューへ
 */
int PlayRandom(uint8_t *RecDirList){
	FILE		*Fil;
    int			ret;
    uint16_t    numdirs;
    uint16_t    numsnds;
    uint16_t    dirno;
    uint8_t    	start_trackno;
    uint32_t    ofs;
    uint16_t    lastdirno;
	
	ret = GOTO_MAINMENU;
	
	Fil = fopen(DIRSTRUCTUREFILE13, "r");
	if (Fil == NULL){
		return 0;
	}
	
    // Read HEADER info
    if (fread(RecDirList, 1, RECLEN_DIRLIST_FULL, Fil) != RECLEN_DIRLIST_FULL){
	    fclose(Fil);
    	return 0;
    }
    memcpy(&numdirs, RecDirList, 2);
	AutoPoffTimer = millis() + AUTOPOFF_PLAYING_TIMEOUT;
	
	while (1){
        if (PonRestart == 1){
            memcpy(&lastdirno, PlayDirHist + HS_LASTDIR_OFFSET, HS_LASTDIR_LEN);
            if (lastdirno >= 1 && lastdirno <= numdirs){
                dirno = lastdirno;
                start_trackno = PlayDirHist[HS_LASTTRACK_OFFSET];
            } else {
                PonRestart = 0;
                continue;
            }
            printf("PonRestart dirno=%d start_trackno=%d ", dirno, start_trackno);
        } else {
            memcpy(&dirno, PlayDirHist + HS_NEXTRANDOMDIR_OFFSET, HS_NEXTRANDOMDIR_LEN);
            if (!(dirno >= 1 && dirno <= numdirs)) dirno = 1;
            start_trackno = 0;
        }
		
        ofs = (uint32_t)(dirno) * (uint32_t)RECLEN_DIRLIST_FULL;
        if (fseek(Fil, ofs, SEEK_SET) == 0){
        	if (fread(RecDirList, 1, RECLEN_DIRLIST_FULL, Fil) == RECLEN_DIRLIST_FULL){
                numsnds = *(RecDirList + DL_NUM_SNDFILE_OFFSET);    // number of sound files
                if (numsnds > 0){
                    if (!(start_trackno >= 0 && start_trackno < numsnds)){
                        start_trackno = 0;
                    }
                    if (PonRestart == 0){
                        memcpy(PlayDirHist + HS_NEXTRANDOMDIR_OFFSET, RecDirList + DL_RECNO_NEXTRANDOMDIR_OFFSET, 2);
                    }
                    //
                    ret = PlayOrSelectMp3InSubdirectory(RecDirList, dirno, numsnds, start_trackno, PS_PLAY);
                    if (ret == POWEROFF){
                        break;
                    } else if (ret == GOTO_MAINMENU){
                        break;
                    }
                } else {
                }
        	}
        }
        PonRestart = 0;
	}
	
    fclose(Fil);
    
    return ret;
}


/*
 * MP3再生
 * 	mp3file:MP3ファイルフルパス
 *	dirno:ディレクトリ構成ファイルのレコード番号（ルートディレクトリ=1から始まる行番号）
 *	trackno:0から始まる曲番号
 *	
 */
int PlayOneMp3(const char *mp3file, uint16_t dirno, uint8_t trackno){
	pid_t	pid;
	int 	status;
	int		fd[2];
	char	fdstr[24];
	char	stereo[24];
	char	attenuate[24];
	char	start[24];
	char	cmd[2];
	char	*args[10];			// madplay起動時引数配列
	int		argidx;
    int		ret;
	unsigned int	start_time_ms;			// 再生開始時刻
	unsigned int	current_time_ms;		// 現在時刻
	unsigned int	notify_time_ms;			// 次回PIC通知時刻
	unsigned int	playtime_time_ms;		// 演奏時間表示次回更新時刻
	unsigned int	lcdoff_time_ms;			// LCD表示自動オフ時刻
	int		restart_time = 0;
	int		lcddispon_flg = 0;
	
restart_label:

	if (PonRestart == 1){       // POWER ON RESTART
		memcpy(&restart_time, PlayDirHist + HS_LASTTIME_OFFSET, HS_LASTTIME_LEN);
		PonRestart = 0;
	} else {
	    memcpy(PlayDirHist + HS_LASTDIR_OFFSET, &dirno, HS_LASTDIR_LEN);
	    PlayDirHist[HS_LASTTRACK_OFFSET] = trackno;
		memcpy(PlayDirHist + HS_LASTTIME_OFFSET, &restart_time, HS_LASTTIME_LEN);
		SaveHistory();
	}
	
	if (socketpair(AF_LOCAL, SOCK_STREAM, 0, fd) == -1){
		printf("socketpair error %d\n", errno);
		return PLAY_NEXT_MP3;
	}
	//printf("fd=%d, %d\n", fd[0], fd[1]);
	
	pid = fork();
	switch(pid){
	case -1:
		exit(1);
		break;
	case 0:
		argidx = 0;
		args[argidx++] = "madplay";
		args[argidx++] = (char *)mp3file;
		sprintf(fdstr, "--socket=%d", fd[1]);
		args[argidx++] = fdstr;
		if (StereoStatus == OUTPUT_MONO){
			sprintf(stereo, "--mono");
		} else {
			sprintf(stereo, "--stereo");
		}
		args[argidx++] = stereo;
		if (Volume < 0){
			sprintf(attenuate, "--attenuate=%d", Volume);
			args[argidx++] = attenuate;
		} else if (Volume > 0){
			sprintf(attenuate, "--amplify=%d", Volume);
			args[argidx++] = attenuate;
		}
		if (restart_time > 0){
			int h, m, s;
			h = (restart_time / 3600);			// hour
			m = (restart_time - h*3600) / 60;	// min
			s = (restart_time % 60);
			sprintf(start, "--start=%02d:%02d:%02d.000", h, m, s);
			args[argidx++] = start;
		}
		args[argidx++] = (char *)0;
		//for (i=0; i<(argidx - 1); i++){
		//	printf("arg=%s\n", args[i]);
		//}
		//execl("/usr/local/bin/madplay", "madplay", mp3file, fdstr, stereo, attenuate, (char *)0);
		execv("/usr/local/bin/madplay", args);
		_exit(errno);
		break;
	default:
		printf("madplay started %d\n", pid);
		start_time_ms = millis();
		notify_time_ms = start_time_ms + NOTIFYPIC_INTERVAL;
		playtime_time_ms = start_time_ms + PLAYTIME_INTERVAL;
		lcdoff_time_ms = start_time_ms + LCDOFF_TIMEOUT;
    	SetPlayTimeInImage((unsigned int)restart_time);
    	DispLcdFullImage();
		ret = 0;
		PinEvent = 0;
		do {
			if (PinEvent != 0 && ret == 0){
				// retの値はMP3_12/decoder.c(run_sync())に合わせる
				if (PinEvent == SWUP_PRESSED){
					cmd[0] = KEY_GAININCR;
					cmd[1] = VOLUME_STEP;
    				write(fd[0], cmd, 2);
    				Volume += VOLUME_STEP;
    				lcddispon_flg = 1;
				} else if (PinEvent == SWDOWN_PRESSED){
					cmd[0] = KEY_GAINDECR;
					cmd[1] = VOLUME_STEP;
    				write(fd[0], cmd, 2);
    				Volume -= VOLUME_STEP;
    				lcddispon_flg = 1;
				} else if (PinEvent == SWNEXT_PRESSED){
					unsigned int current_play_time = (unsigned int)((current_time_ms - start_time_ms)/1000) + restart_time;
					if (SwNextAtPlaying(&current_play_time) == 0){
	    				ret = PLAY_NEXT_MP3;
					} else {
						restart_time = current_play_time;
						ret = PLAY_ONCEMORE;
					}
					cmd[0] = KEY_QUIT;
    				write(fd[0], cmd, 1);
				} else if (PinEvent == SWPREV_PRESSED){
					unsigned int current_play_time = (unsigned int)((current_time_ms - start_time_ms)/1000) + restart_time;
					if (SwPrevAtPlaying(&current_play_time) == 0){
	   					ret = PLAY_PREV_MP3;
	   				} else {
						restart_time = current_play_time;
						ret = PLAY_ONCEMORE;
	   				}
					cmd[0] = KEY_QUIT;
    				write(fd[0], cmd, 1);
				} else if (PinEvent == SWPLAY_PRESSED){
					cmd[0] = KEY_QUIT;
    				write(fd[0], cmd, 1);
    				ret = PLAY_OTHER_DIR;
				} else if (PinEvent == SWDESELECT_PRESSED){
					cmd[0] = KEY_QUIT;
    				write(fd[0], cmd, 1);
    				ret = SELECT_OTHER_MP3;
				} else if (PinEvent == SWOPTION_PRESSED){
					lcddispon_flg = 1;
				} else if (PinEvent == PDOWN_PRESSED){
					cmd[0] = KEY_QUIT;
    				write(fd[0], cmd, 1);
    				ret = POWEROFF;
					break;
				}
				if (PinEvent != 0){
					printf("PinEvent=%d\n", PinEvent);
				}
				PinEvent = 0;
			}
			current_time_ms = millis();
			if (current_time_ms > AutoPoffTimer){
				cmd[0] = KEY_QUIT;
				write(fd[0], cmd, 1);
				ret = POWEROFF;
				break;		// 明示的にbreakしないと無限ループになる？
			} else if (lcddispon_flg == 1 || 
					(LcdDispStatus == 1 && current_time_ms > playtime_time_ms)){
				SetPlayTimeInImage((unsigned int)((current_time_ms - start_time_ms)/1000) + restart_time);
				DispLcdFullImage();
				usleep(200*1000);	// 200ms (PICのLCD出力を待つ)
				playtime_time_ms = current_time_ms + PLAYTIME_INTERVAL;
				if (lcddispon_flg == 1){
					lcdoff_time_ms = current_time_ms + LCDOFF_TIMEOUT;
					lcddispon_flg = 0;
				}
			} else if (current_time_ms > notify_time_ms){
				notifyactive(FdSpi);		// 本PGが実行中であることを定期的にPICに通知
				notify_time_ms = current_time_ms + NOTIFYPIC_INTERVAL;
			}
			if (lcdoff_time_ms > 0 && current_time_ms > lcdoff_time_ms){
				lcddispoff(FdSpi);
				lcdoff_time_ms = 0;
			}
			
			// madplayの実行状態
			if ((pid = waitpid(pid, &status, WNOHANG)) == -1){
				printf("waitpid error");
				break;
			} else if (pid != 0) {
				//sleep(1);
			//}
			//else {
				if (WIFEXITED(status))
					printf("child exited with status of %d¥n", WEXITSTATUS(status));
				else puts("child did not exit successfully");
			}
		} while (pid == 0);
		close(fd[0]);
		close(fd[1]);
		
		printf("madplay end\n");
		
		break;
	}
	
	if (ret == PLAY_ONCEMORE){
		goto restart_label;
	} else {
		restart_time += (millis() - start_time_ms)/1000;		// 経過時間(秒)
		memcpy(PlayDirHist + HS_LASTTIME_OFFSET, &restart_time, HS_LASTTIME_LEN);
		SaveHistory();
	}
	
	return ret;
}


/*
 *	再生中にNEXTボタンが押されたときの動作
 *	current_play_time : 現在の再生時間（秒）＝ 曲の最初からの経過時間
 *	長押しされたとき戻り値 1、通常押しの戻り値 0
 */
int SwNextAtPlaying(unsigned int *current_play_time){
	unsigned int	start_time_ms;
	unsigned int	current_time_ms;
	int				ret = 0;
	unsigned int	step_time = 10;		// sec
	int				cnt = 0;
	
	SetPlayTimeInImage(*current_play_time);
	DispLcdFullImage();
	start_time_ms = millis();
	while (digitalRead(PIN_SWNEXT) == 0){
		current_time_ms = millis();
		if (current_time_ms - start_time_ms > LONGPRESS_INTERVAL){
			ret = 1;
			*current_play_time += step_time;
			SetPlayTimeInImage(*current_play_time);
			DispLcdFullImage();
			start_time_ms = current_time_ms;
			cnt++;
			if (cnt == 5) step_time = 60;			// 1min
			else if (cnt == 10) step_time = 120;	// 2min
			else if (cnt == 15) step_time = 300;	// 5min
		}
	}
	usleep(200*1000);	// 200ms (PICのLCD出力を待つ)
	return ret;
}


/*
 *	再生中にPREVボタンが押されたときの動作
 *	current_play_time : 現在の再生時間（秒）＝ 曲の最初からの経過時間
 *	長押しされたとき戻り値 1、通常押しの戻り値 0
 */
int SwPrevAtPlaying(unsigned int *current_play_time){
	unsigned int	start_time_ms;
	unsigned int	current_time_ms;
	int				ret = 0;
	unsigned int	step_time = 10;		// sec
	int				cnt = 0;
	
	SetPlayTimeInImage(*current_play_time);
	DispLcdFullImage();
	start_time_ms = millis();
	while (digitalRead(PIN_SWPREV) == 0){
		current_time_ms = millis();
		if (current_time_ms - start_time_ms > LONGPRESS_INTERVAL){
			ret = 1;
			if (*current_play_time > step_time){
				*current_play_time -= step_time;
			} else {
				*current_play_time = 0;
			}
			SetPlayTimeInImage(*current_play_time);
			DispLcdFullImage();
			start_time_ms = current_time_ms;
			cnt++;
			if (cnt == 5) step_time = 60;			// 1min
			else if (cnt == 10) step_time = 120;	// 2min
			else if (cnt == 15) step_time = 300;	// 5min
		}
	}
	usleep(200*1000);	// 200ms (PICのLCD出力を待つ)
	return ret;
}


/*
 * 再生中の表示イメージ
 */
void SetImageAtPlaying(FILE *Fil, uint8_t *RecMp3List, uint8_t trackno, uint16_t numsndfile, uint8_t NumSelItem,
		uint8_t *RecDirList){
	uint8_t		start_trackno;
	uint8_t		t;

	if ((trackno - 1) >= 0){
		start_trackno = trackno - 1;
	} else {
		start_trackno = 0;
	}

	memset(LcdImageBuff, 0, LCD_IMAGE_BUFFER_SIZE);
	// １行目：ディレクトリ名
	SetRecDirList2MenuImage(RecDirList + RECLEN_DIRLIST, 0, NumSelItem);
	
	// ２行目から：曲名
	fseek(Fil, (uint32_t)(start_trackno)*(RECLEN_MP3LIST_FULL), SEEK_SET);
	for (t=start_trackno; (t - start_trackno)<(NumSelItem-1) && t<numsndfile; t++){
       	if (fread(RecMp3List, 1, RECLEN_MP3LIST_FULL, Fil) == RECLEN_MP3LIST_FULL){
			SetRecDirList2MenuImage(RecMp3List + RECLEN_MP3LIST, t - start_trackno + 1, NumSelItem);
		}
	}
	// 再生中の曲を反転表示
	SetMenuItemSelected(trackno - start_trackno + 1, NumSelItem);
}


/*
 *	イメージデータに演奏時間のイメージを埋め込む
 *	play_time : 秒
 */
static void SetPlayTimeInImage(unsigned int play_time){
	unsigned short	img[PLAYTIME_WIDTH];
	unsigned int sec;
	int	d, x_to, x_from, i;
	
	sec = (play_time % 60);
	play_time = (play_time - sec) / 60;
	
	// 1s
	d = (sec % 10);
	x_to = 5 * HAN_WIDTH;
	x_from = d * HAN_WIDTH;
	for (i=0; i<HAN_WIDTH; i++){
		img[x_to + i] = DigitImage[x_from + i];
	}
	
	// 10s
	sec = (sec - d) / 10;
	d = (sec % 10);
	x_to = 4 * HAN_WIDTH;
	x_from = d * HAN_WIDTH;
	for (i=0; i<HAN_WIDTH; i++){
		img[x_to + i] = DigitImage[x_from + i];
	}
	
	// :
	x_to = 3 * HAN_WIDTH;
	x_from = 10 * HAN_WIDTH;		// 数字イメージファイル中 ':'の位置
	for (i=0; i<HAN_WIDTH; i++){
		img[x_to + i] = DigitImage[x_from + i];
	}
	
	// 1m
	d = (play_time % 10);
	x_to = 2 * HAN_WIDTH;
	x_from = d * HAN_WIDTH;
	for (i=0; i<HAN_WIDTH; i++){
		img[x_to + i] = DigitImage[x_from + i];
	}
	
	// 10m
	play_time = (play_time - d) / 10;
	d = (play_time % 10);
	x_to = 1 * HAN_WIDTH;
	if (d > 0 || (d == 0 && play_time >= 10)){
		x_from = d * HAN_WIDTH;
		for (i=0; i<HAN_WIDTH; i++){
			img[x_to + i] = DigitImage[x_from + i];
		}
	} else {
		for (i=0; i<HAN_WIDTH; i++){
			img[x_to + i] = 0;
		}
	}
	
	// 100m
	play_time = (play_time - d) / 10;
	d = (play_time % 10);
	x_to = 0 * HAN_WIDTH;
	if (d > 0){
		x_from = d * HAN_WIDTH;
		for (i=0; i<HAN_WIDTH; i++){
			img[x_to + i] = DigitImage[x_from + i];
		}
	} else {
		for (i=0; i<HAN_WIDTH; i++){
			img[x_to + i] = 0;
		}
	}
	
	x_to = LCD_WIDTH - PLAYTIME_WIDTH;
	for (i=0; i<PLAYTIME_WIDTH; i++){
		LcdImageBuff[6*LCD_WIDTH + x_to] = (img[i] << 4) & 0xF0;	// Page6
		LcdImageBuff[7*LCD_WIDTH + x_to] = (img[i] >> 4) & 0xFF;	// Page7
		x_to++;
	}
	
	x_to = LCD_WIDTH - PLAYTIME_WIDTH - 1;
	LcdImageBuff[6*LCD_WIDTH + x_to] = 0;	// Page6
	LcdImageBuff[7*LCD_WIDTH + x_to] = 0;	// Page7
}


// 数字イメージファイル
void ReadDigitImage(){
	FILE		*Fil;

	Fil = fopen(DIGITIMAGEDAT, "r");
	if (Fil != NULL){
		fread(&DigitImage, DIGITIMAGE_LEN, 1, Fil);
		fclose(Fil);
	}
}

/*
 * 履歴ファイルを読み込む
 */
void ReadHistory(){
	FILE		*Fil;

	Fil = fopen(HISTFILE, "r");
	if (Fil != NULL){
		fread(&PlayDirHist, RECLEN_HISTFILE, 1, Fil);
		fclose(Fil);
		StereoStatus = PlayDirHist[HS_STEREOSTATUS_OFFSET];
		if (StereoStatus != OUTPUT_STEREO && StereoStatus != OUTPUT_MONO){
			StereoStatus = OUTPUT_MONO;
		}
	}
}

/*
 * 履歴ファイルを保存
 */
void SaveHistory(){
	FILE		*Fil;
	
	Fil = fopen(HISTFILE, "w");
	if (Fil != NULL){
		PlayDirHist[HS_STEREOSTATUS_OFFSET] = StereoStatus;
		fwrite(&PlayDirHist, RECLEN_HISTFILE, 1, Fil);
		fclose(Fil);
	}
	
}

// R-CHの電源制御
void SetRchPower(){
	if (StereoStatus == OUTPUT_STEREO){
		digitalWrite(PIN_OUT_PWR_RCH, 1);
	} else {
		digitalWrite(PIN_OUT_PWR_RCH, 0);
	}
}

